import sys
from enum import IntEnum
from functools import wraps
from contextlib import contextmanager

sys.tracebacklimit = 0


class DataBaseError(Exception):
    """Is raised when the DB raises an OperationalError."""

    def __init__(self, message) -> None:
        self.message = message
        super().__init__(message)


class UserInputError(Exception):
    """Is raised when the user inputs invalid arguments."""

    def __init__(self, message) -> None:
        self.message = message
        super().__init__(message)


class FileContentError(Exception):
    """Is raised when file content doesn't have the expected format."""

    def __init__(self, message) -> None:
        self.message = message
        super().__init__(message)


class InvalidDatasetError(Exception):
    """Is raised when a dataset violates the constraints imposed by the data model."""

    def __init__(self, message) -> None:
        self.message = message
        super().__init__(message)


class InvalidDataModelError(Exception):
    """Is raised when the data model doesn't have the expected schema."""

    def __init__(self, message) -> None:
        self.message = message
        super().__init__(message)


class ForeignKeyError(Exception):
    """Is raised when a table is deleted while there is a foreign key constrain from another table."""

    def __init__(self, message) -> None:
        self.message = message
        super().__init__(message)


class ExitCode(IntEnum):
    OK = 0
    USER_ERROR = 64
    DB_ERROR = 65
    FILE_ERROR = 66


def handle_errors(func):
    @contextmanager
    def _handle_errors():
        try:
            yield
        except UserInputError as exc:
            print("User input error:\n")
            print(f"\t{exc.message}")
            sys.exit(ExitCode.USER_ERROR)
        except DataBaseError as exc:
            print("Database error:\n")
            print(f"\t{exc.message}")
            sys.exit(ExitCode.DB_ERROR)
        except FileContentError as exc:
            print("File error:\n")
            print(f"\t{exc.message}")
            sys.exit(ExitCode.FILE_ERROR)
        except ForeignKeyError as exc:
            print("Foreign key error:\n")
            print(f"\t{exc.message}")
        except InvalidDatasetError as exc:
            print(f"\nDataset error: {exc.message}")
            sys.exit(ExitCode.FILE_ERROR)
        # except Exception as exc:
        #     print("Something went wrong:\n")
        #     raise exc

    @wraps(func)
    def wrapper(*args, **kwargs):
        with _handle_errors():
            return func(*args, **kwargs)

    return wrapper



import ipaddress
from pathlib import Path

import click as cl
import os
import glob

from mipdb.databases import credentials_from_config
from mipdb.databases.monetdb import MonetDB
from mipdb.reader import JsonFileReader
from mipdb.databases.sqlite import SQLiteDB
from mipdb.usecases import (
    AddDataModel,
    Cleanup,
    ValidateDatasetNoDatabase,
    ValidateDataModel,
)
from mipdb.usecases import AddPropertyToDataModel
from mipdb.usecases import AddPropertyToDataset
from mipdb.usecases import DeleteDataModel
from mipdb.usecases import DeleteDataset
from mipdb.usecases import ImportCSV
from mipdb.usecases import InitDB
from mipdb.exceptions import handle_errors, UserInputError
from mipdb.usecases import DisableDataset
from mipdb.usecases import DisableDataModel
from mipdb.usecases import EnableDataset
from mipdb.usecases import EnableDataModel
from mipdb.usecases import ListDataModels
from mipdb.usecases import ListDatasets
from mipdb.usecases import RemovePropertyFromDataModel
from mipdb.usecases import RemovePropertyFromDataset
from mipdb.usecases import UntagDataModel
from mipdb.usecases import TagDataModel
from mipdb.usecases import TagDataset
from mipdb.usecases import UntagDataset
from mipdb.usecases import ValidateDataset


class NotRequiredIf(cl.Option):
    def __init__(self, *args, **kwargs):
        credentials = credentials_from_config()
        option_to_env_var = {
            "--ip": credentials["DB_IP"],
            "--port": credentials["DB_PORT"],
            "--username": credentials["MONETDB_ADMIN_USERNAME"],
            "--password": credentials["MONETDB_LOCAL_PASSWORD"],
            "--db_name": credentials["DB_NAME"],
            "--sqlite_db_path": credentials["SQLITE_DB_PATH"],
        }
        option = args[0][0]
        if option_to_env_var[option]:
            kwargs["required"] = False
            kwargs["default"] = option_to_env_var[option]
        super(NotRequiredIf, self).__init__(*args, **kwargs)


_db_configs_options = [
    cl.option(
        "--ip",
        "ip",
        required=True,
        help="The ip of the database",
        cls=NotRequiredIf,
    ),
    cl.option(
        "--port",
        "port",
        required=True,
        help="The port for the database",
        cls=NotRequiredIf,
    ),
    cl.option(
        "--username",
        "username",
        required=True,
        help="The username for the database",
        cls=NotRequiredIf,
    ),
    cl.option(
        "--password",
        "password",
        required=True,
        help="The password for the database",
        cls=NotRequiredIf,
    ),
    cl.option(
        "--db_name",
        "db_name",
        required=True,
        help="The name of the database",
        cls=NotRequiredIf,
    ),
    cl.option(
        "--sqlite_db_path",
        "sqlite_db_path",
        required=True,
        help="The path for the sqlite database",
        cls=NotRequiredIf,
    ),
]


def get_monetdb_config(ip, port, username, password, db_name):
    try:
        ipaddress.ip_address(ip)
    except ValueError:
        raise UserInputError("Invalid ip provided")

    config = {
        "ip": ip,
        "port": port,
        "dbfarm": db_name,
        "username": username,
        "password": password,
    }
    return config


def db_configs_options(func):
    for option in reversed(_db_configs_options):
        func = option(func)
    return func


@cl.group()
def entry():
    pass


@entry.command()
@cl.argument("file", required=True)
@cl.option(
    "--copy_from_file",
    required=False,
    default=True,
    help="Copy the csvs from the filesystem instead of copying them through sockets."
    "The same files should exist both in the mipdb script and the db.",
)
@db_configs_options
@handle_errors
def load_folder(
    file, copy_from_file, ip, port, username, password, db_name, sqlite_db_path
):
    dbconfig = get_monetdb_config(ip, port, username, password, db_name)
    monetdb = MonetDB.from_config(dbconfig)
    sqlite_db = SQLiteDB.from_config({"db_path": sqlite_db_path})

    Cleanup(sqlite_db, monetdb).execute()
    if not os.path.exists(file):
        print(f"The path {file} does not exist.")
        return
    elif not os.listdir(file):
        print(f"The directory {file} is empty.")
        return

    root = Path(file).resolve()
    for subdir, dirs, files in os.walk(file):
        if dirs:
            continue
        print(f"Data model '{subdir}' is being loaded...")
        metadata_path = os.path.join(subdir, "CDEsMetadata.json")
        reader = JsonFileReader(metadata_path)
        data_model_metadata = reader.read()
        code = data_model_metadata["code"]
        version = data_model_metadata["version"]
        AddDataModel(sqlite_db=sqlite_db, monetdb=monetdb).execute(data_model_metadata)
        print(f"Data model '{code}' was successfully added.")

        for csv_path in glob.glob(subdir + "/*.csv"):
            print(f"CSV '{csv_path}' is being loaded...")
            ValidateDataset(sqlite_db=sqlite_db, monetdb=monetdb).execute(
                csv_path, copy_from_file, code, version
            )
            ImportCSV(sqlite_db=sqlite_db, monetdb=monetdb).execute(
                csv_path, copy_from_file, code, version
            )
            print(f"CSV '{csv_path}' was successfully added.")


@entry.command()
@cl.argument("file", required=True)
@handle_errors
def validate_folder(file):
    if not os.path.exists(file):
        print(f"The path {file} does not exist.")
        return
    elif not os.listdir(file):
        print(f"The directory {file} is empty.")
        return

    for subdir, dirs, files in os.walk(file):
        if dirs:
            continue
        print(f"Data model '{subdir}' is being validated...")
        metadata_path = os.path.join(subdir, "CDEsMetadata.json")
        reader = JsonFileReader(metadata_path)
        data_model_metadata = reader.read()
        code = data_model_metadata["code"]
        ValidateDataModel().execute(data_model_metadata)
        print(f"Data model '{code}' was successfully validated.")

        for csv_path in glob.glob(subdir + "/*.csv"):
            print(f"CSV '{csv_path}' is being validated...")
            ValidateDatasetNoDatabase().execute(csv_path, data_model_metadata)
            print(f"CSV '{csv_path}' was successfully validated.")


@entry.command()
@cl.option(
    "--sqlite_db_path",
    "sqlite_db_path",
    required=True,
    help="The path for the sqlite database",
    cls=NotRequiredIf,
)
@handle_errors
def init(sqlite_db_path):
    sqlite_db = SQLiteDB.from_config({"db_path": sqlite_db_path})
    InitDB(db=sqlite_db).execute()
    print("Database initialized")


@entry.command()
@cl.argument("file", required=True)
@db_configs_options
@handle_errors
def add_data_model(file, ip, port, username, password, db_name, sqlite_db_path):
    print(f"Data model '{file}' is being loaded...")
    dbconfig = get_monetdb_config(ip, port, username, password, db_name)
    reader = JsonFileReader(file)
    monetdb = MonetDB.from_config(dbconfig)
    sqlite_db = SQLiteDB.from_config({"db_path": sqlite_db_path})
    data_model_metadata = reader.read()
    ValidateDataModel().execute(data_model_metadata)
    AddDataModel(sqlite_db=sqlite_db, monetdb=monetdb).execute(data_model_metadata)
    print(f"Data model '{file}' was successfully added.")


@entry.command()
@cl.argument("csv_path", required=True)
@cl.option(
    "-d",
    "--data-model",
    required=True,
    help="The data model to which the dataset is added",
)
@cl.option("-v", "--version", required=True, help="The data model version")
@cl.option(
    "--copy_from_file",
    required=False,
    default=True,
    help="Copy the csvs from the filesystem instead of copying them through sockets."
    "The same files should exist both in the mipdb script and the db.",
)
@db_configs_options
@handle_errors
def add_dataset(
    csv_path,
    data_model,
    version,
    copy_from_file,
    ip,
    port,
    username,
    password,
    db_name,
    sqlite_db_path,
):
    print(f"CSV '{csv_path}' is being loaded...")
    dbconfig = get_monetdb_config(ip, port, username, password, db_name)
    monetdb = MonetDB.from_config(dbconfig)
    sqlite_db = SQLiteDB.from_config({"db_path": sqlite_db_path})
    ValidateDataset(sqlite_db=sqlite_db, monetdb=monetdb).execute(
        csv_path, copy_from_file, data_model, version
    )
    ImportCSV(sqlite_db=sqlite_db, monetdb=monetdb).execute(
        csv_path, copy_from_file, data_model, version
    )
    print(f"CSV '{csv_path}' was successfully added.")


@entry.command()
@cl.argument("csv_path", required=True)
@cl.option(
    "-d",
    "--data-model",
    required=True,
    help="The data model to which the dataset is added",
)
@cl.option("-v", "--version", required=True, help="The data model version")
@cl.option(
    "--copy_from_file",
    required=False,
    default=True,
    help="Copy the csvs from the filesystem instead of copying them through sockets."
    "The same files should exist both in the mipdb script and the db.",
)
@db_configs_options
@handle_errors
def validate_dataset(
    csv_path,
    data_model,
    version,
    copy_from_file,
    ip,
    port,
    username,
    password,
    db_name,
    sqlite_db_path,
):
    print(f"Dataset '{csv_path}' is being validated...")
    dbconfig = get_monetdb_config(ip, port, username, password, db_name)
    monetdb = MonetDB.from_config(dbconfig)
    sqlite_db = SQLiteDB.from_config({"db_path": sqlite_db_path})

    ValidateDataset(sqlite_db=sqlite_db, monetdb=monetdb).execute(
        csv_path, copy_from_file, data_model, version
    )
    print(f"Dataset '{csv_path}' has a valid structure.")


@entry.command()
@cl.argument("name", required=True)
@cl.option("-v", "--version", required=True, help="The data model version")
@cl.option(
    "--force",
    "-f",
    is_flag=True,
    help="Force deletion of dataset that are based on the data model",
)
@db_configs_options
@handle_errors
def delete_data_model(
    name, version, force, ip, port, username, password, db_name, sqlite_db_path
):
    dbconfig = get_monetdb_config(ip, port, username, password, db_name)
    monetdb = MonetDB.from_config(dbconfig)
    sqlite_db = SQLiteDB.from_config({"db_path": sqlite_db_path})
    DeleteDataModel(sqlite_db=sqlite_db, monetdb=monetdb).execute(name, version, force)
    print(f"Data model '{name}' was successfully removed.")


@entry.command()
@cl.argument("dataset", required=True)
@cl.option(
    "-d",
    "--data-model",
    required=True,
    help="The data model to which the dataset is added",
)
@cl.option("-v", "--version", required=True, help="The data model version")
@db_configs_options
@handle_errors
def delete_dataset(
    dataset, data_model, version, ip, port, username, password, db_name, sqlite_db_path
):
    dbconfig = get_monetdb_config(ip, port, username, password, db_name)
    monetdb = MonetDB.from_config(dbconfig)
    sqlite_db = SQLiteDB.from_config({"db_path": sqlite_db_path})
    DeleteDataset(sqlite_db=sqlite_db, monetdb=monetdb).execute(
        dataset, data_model, version
    )
    print(f"Dataset {dataset} was successfully removed.")


@entry.command()
@cl.argument("name", required=True)
@cl.option("-v", "--version", required=True, help="The data model version")
@cl.option(
    "--sqlite_db_path",
    "sqlite_db_path",
    required=True,
    help="The path for the sqlite database",
    cls=NotRequiredIf,
)
@handle_errors
def enable_data_model(name, version, sqlite_db_path):
    sqlite_db = SQLiteDB.from_config({"db_path": sqlite_db_path})
    EnableDataModel(db=sqlite_db).execute(name, version)
    print(f"Data model {name} was successfully enabled.")


@entry.command()
@cl.argument("name", required=True)
@cl.option("-v", "--version", required=True, help="The data model version")
@cl.option(
    "--sqlite_db_path",
    "sqlite_db_path",
    required=True,
    help="The path for the sqlite database",
    cls=NotRequiredIf,
)
@handle_errors
def disable_data_model(name, version, sqlite_db_path):
    sqlite_db = SQLiteDB.from_config({"db_path": sqlite_db_path})
    DisableDataModel(db=sqlite_db).execute(name, version)
    print(f"Data model {name} was successfully disabled.")


@entry.command()
@cl.argument("dataset", required=True)
@cl.option(
    "-d",
    "--data-model",
    required=True,
    help="The data model to which the dataset is added",
)
@cl.option("-v", "--version", required=True, help="The data model version")
@cl.option(
    "--sqlite_db_path",
    "sqlite_db_path",
    required=True,
    help="The path for the sqlite database",
    cls=NotRequiredIf,
)
@handle_errors
def enable_dataset(dataset, data_model, version, sqlite_db_path):
    sqlite_db = SQLiteDB.from_config({"db_path": sqlite_db_path})
    EnableDataset(db=sqlite_db).execute(dataset, data_model, version)
    print(f"Dataset {dataset} was successfully enabled.")


@entry.command()
@cl.argument("dataset", required=True)
@cl.option(
    "-d",
    "--data-model",
    required=True,
    help="The data model to which the dataset is added",
)
@cl.option("-v", "--version", required=True, help="The data model version")
@cl.option(
    "--sqlite_db_path",
    "sqlite_db_path",
    required=True,
    help="The path for the sqlite database",
    cls=NotRequiredIf,
)
@handle_errors
def disable_dataset(dataset, data_model, version, sqlite_db_path):
    sqlite_db = SQLiteDB.from_config({"db_path": sqlite_db_path})
    DisableDataset(sqlite_db).execute(dataset, data_model, version)
    print(f"Dataset {dataset} was successfully disabled.")


@entry.command()
@cl.argument("name", required=True)
@cl.option("-v", "--version", required=True, help="The data model version")
@cl.option(
    "-t",
    "--tag",
    default=None,
    required=True,
    help="A tag to be added/removed",
)
@cl.option(
    "-r",
    "--remove",
    is_flag=True,
    required=False,
    help="A flag that determines if the tag/key_value will be removed",
)
@cl.option(
    "--force",
    "-f",
    is_flag=True,
    help="Force overwrite on property",
)
@cl.option(
    "--sqlite_db_path",
    "sqlite_db_path",
    required=True,
    help="The path for the sqlite database",
    cls=NotRequiredIf,
)
@handle_errors
def tag_data_model(name, version, tag, remove, force, sqlite_db_path):
    sqlite_db = SQLiteDB.from_config({"db_path": sqlite_db_path})
    if "=" in tag:
        key, value = tag.split("=")
        if remove:
            RemovePropertyFromDataModel(db=sqlite_db).execute(name, version, key, value)
            print(f"Property was successfully removed from data model {name}.")
        else:
            AddPropertyToDataModel(db=sqlite_db).execute(
                name, version, key, value, force
            )
            print(f"Property was successfully added to data model {name}.")
    else:
        if remove:
            UntagDataModel(db=sqlite_db).execute(name, version, tag)
            print(f"Data model {name} was successfully untagged.")
        else:
            TagDataModel(db=sqlite_db).execute(name, version, tag)
            print(f"Data model {name} was successfully tagged.")


@entry.command()
@cl.argument("dataset", required=True)
@cl.option(
    "-d",
    "--data-model",
    required=True,
    help="The data model to which the dataset is added",
)
@cl.option("-v", "--version", required=True, help="The data model version")
@cl.option(
    "-t",
    "--tag",
    default=None,
    required=True,
    help="A tag to be added/removed",
)
@cl.option(
    "-r",
    "--remove",
    is_flag=True,
    required=False,
    help="A flag that determines if the tag/key_value will be removed",
)
@cl.option(
    "--force",
    "-f",
    is_flag=True,
    help="Force overwrite on property",
)
@cl.option(
    "--sqlite_db_path",
    "sqlite_db_path",
    required=True,
    help="The path for the sqlite database",
    cls=NotRequiredIf,
)
@handle_errors
def tag_dataset(
    dataset,
    data_model,
    version,
    tag,
    remove,
    force,
    sqlite_db_path,
):
    sqlite_db = SQLiteDB.from_config({"db_path": sqlite_db_path})

    if "=" in tag:
        key, value = tag.split("=")
        if remove:
            RemovePropertyFromDataset(db=sqlite_db).execute(
                dataset, data_model, version, key, value
            )
            print(f"Property was successfully removed from dataset {dataset}.")
        else:
            AddPropertyToDataset(db=sqlite_db).execute(
                dataset, data_model, version, key, value, force
            )
            print(f"Property was successfully added to dataset {dataset}.")
    else:
        if remove:
            UntagDataset(db=sqlite_db).execute(dataset, data_model, version, tag)
            print(f"Dataset {dataset} was successfully untagged.")
        else:
            TagDataset(db=sqlite_db).execute(dataset, data_model, version, tag)
            print(f"Dataset {dataset} was successfully tagged.")


@entry.command()
@cl.option(
    "--sqlite_db_path",
    "sqlite_db_path",
    required=True,
    help="The path for the sqlite database",
    cls=NotRequiredIf,
)
@handle_errors
def list_data_models(sqlite_db_path):
    sqlite_db = SQLiteDB.from_config({"db_path": sqlite_db_path})
    ListDataModels(db=sqlite_db).execute()


@entry.command()
@db_configs_options
@handle_errors
def list_datasets(ip, port, username, password, db_name, sqlite_db_path):
    dbconfig = get_monetdb_config(ip, port, username, password, db_name)
    monetdb = MonetDB.from_config(dbconfig)
    sqlite_db = SQLiteDB.from_config({"db_path": sqlite_db_path})
    ListDatasets(sqlite_db=sqlite_db, monetdb=monetdb).execute()



import warnings
from abc import ABC, abstractmethod

import sqlalchemy as sql
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Integer, String, JSON,  select

from typing import List

from mipdb.dataelements import CommonDataElement
from mipdb.exceptions import DataBaseError
from mipdb.databases.sqlite import DataModel, Dataset

METADATA_TABLE = "variables_metadata"
PRIMARYDATA_TABLE = "primary_data"

Base = declarative_base()


class Status:
    ENABLED = "ENABLED"
    DISABLED = "DISABLED"


class SQLTYPES:
    INTEGER = Integer
    STRING = String(255)
    FLOAT = sql.Float
    JSON = JSON


def get_metadata_table_name(data_model):
    return f"{data_model}_{METADATA_TABLE}"


def handle_errors(func):
    """Decorator for any function susceptible to raise a DB related exception."""

    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as exc:  # Use generic Exception to capture all errors
            raise DataBaseError(f"Database error: {exc}")

    return wrapper


class Table(ABC):
    _table: Base

    @abstractmethod
    def __init__(self):
        pass

    @property
    def table(self):
        return self._table

    def create(self, db):
        db.create_table(self._table)

    def exists(self, db):
        return db.table_exists(self._table)

    def insert_values(self, values, db):
        db.insert_values_to_table(self._table, values)

    def delete(self, db):
        db.delete_from(self._table, where_conditions={})

    def get_row_count(self, db):
        return db.get_row_count(self.table.fullname)

    def get_column_distinct(self, column, db):
        return db.get_column_distinct(column, self.table.fullname)

    def drop(self, db):
        db.drop_table(self._table)

class DataModelTable(Table):
    def __init__(self):
        self._table = DataModel.__table__

    def get_data_models(self, db, columns: list = None):
        return db.get_values(table=self._table, columns=columns)

    def get_data_model(self, data_model_id, db, columns: list = None):
        return db.get_data_models_values(
            columns=columns, where_conditions={"data_model_id": data_model_id}
        )

    def get_dataset_count_by_data_model_id(self, db):
        return db.get_dataset_count_by_data_model_id()

    def get_data_model_id(self, code, version, db):
        return db.get_data_model_id(code, version)

    def get_data_model_properties(self, data_model_id, db):
        return db.get_data_model_properties(data_model_id)

    def set_data_model_properties(self, properties, data_model_id, db):
        db.set_data_model_properties(properties, data_model_id)

    def get_data_model_status(self, data_model_id, db):
        return db.get_data_model_status(data_model_id)

    def set_data_model_status(self, status, data_model_id, db):
        db.update_data_model_status(status, data_model_id)

    def delete_data_model(self, code, version, db):
        db.delete_from(self._table, where_conditions={"code": code, "version": version})

    def get_next_data_model_id(self, db) -> int:
        res = db.execute_fetchall('SELECT MAX(data_model_id) FROM data_models;')
        max_id = res[0][0] if res and res[0][0] is not None else 0
        return max_id + 1


class DatasetsTable(Table):
    def __init__(self):
        self._table = Dataset.__table__

    def get_datasets(self, db, columns: list = None):
        return db.get_values(table=self._table, columns=columns, where_conditions={})

    def get_dataset_codes(self, db, columns: List[str] = None, data_model_id: int = None) -> List[str]:
        cols = columns or ['code']
        stmt = select(*[self.table.c[col] for col in cols])
        if data_model_id is not None:
            stmt = stmt.where(self.table.c.data_model_id == data_model_id)
        compiled_sql = str(
            stmt.compile(compile_kwargs={"literal_binds": True})
        )
        rows = db.execute_fetchall(compiled_sql)
        codes: List[str] = []
        for row in rows:
            try:
                codes.append(row[0])
            except Exception:
                codes.append(row.get(cols[0]))
        return codes

    def get_dataset(self, db, dataset_id=None, columns=None):
        return db.get_values(
            table=self._table,
            columns=columns,
            where_conditions={"dataset_id": dataset_id},
        )

    def get_dataset_properties(self, dataset_id, db):
        return db.get_dataset_properties(dataset_id)

    def set_dataset_properties(self, properties, dataset_id, db):
        db.set_dataset_properties(properties, dataset_id)

    def delete_dataset(self, dataset_id, data_model_id, db):
        db.delete_from(
            self._table,
            where_conditions={"dataset_id": dataset_id, "data_model_id": data_model_id},
        )

    def get_next_dataset_id(self, db):
        result = db.get_max_dataset_id()
        return result + 1 if result else 1

    def get_dataset_status(self, data_model_id, db):
        return db.get_dataset_status(data_model_id)

    def set_dataset_status(self, status, dataset_id, db):
        db.update_dataset_status(status, dataset_id)

    def get_dataset_id(self, code, data_model_id, db):
        return db.get_dataset_id(code, data_model_id)


class MetadataTable(Table):
    def __init__(self, data_model):
        self.name = get_metadata_table_name(data_model)
        self._table = sql.Table(
            self.name,
            Base.metadata,
            sql.Column("code", SQLTYPES.STRING, primary_key=True),
            sql.Column("metadata", SQLTYPES.JSON),
            extend_existing=True,
        )

    def set_table(self, table):
        self._table = table

    @classmethod
    def from_db(cls, data_model, db):
        res = db.get_metadata(data_model)
        new_table = cls(data_model)
        new_table.set_table(
            {
                code: CommonDataElement.from_metadata(metadata)
                for code, metadata in res.items()
            }
        )
        return new_table

    @staticmethod
    def get_values_from_cdes(cdes):
        return [{"code": cde.code, "metadata": cde.metadata} for cde in cdes]

    def insert_values(self, values, db):
        db.execute(f'INSERT INTO "{self.name}" VALUES(:code, :metadata)', values)



import pandas as pd
import pandera.pandas as pa

from mipdb.exceptions import InvalidDatasetError


class DataFrameSchema:
    _schema: pa.DataFrameSchema

    def __init__(
        self, sql_type_per_column, cdes_with_min_max, cdes_with_enumerations, columns
    ) -> None:
        pa_columns = {}
        # Validating the dataset has proper values, according to the data model.

        # There is a need to construct a DataFrameSchema with all the constraints that the metadata is imposing
        # For each column a pandera Column is created that will contain the constraints for the specific column
        if not set(columns) <= set(sql_type_per_column.keys()):
            raise InvalidDatasetError(
                f"Columns:{set(columns) - set(sql_type_per_column.keys()) - {'row_id'} } are not present in the CDEs"
            )

        for column in columns:
            checks = self._get_pa_checks(
                cdes_with_min_max, cdes_with_enumerations, column
            )
            cde_sql_type = sql_type_per_column[column]
            pa_type = self._pa_type_from_sql_type(cde_sql_type)
            pa_columns[column] = pa.Column(dtype=pa_type, checks=checks, nullable=True)

        self._schema = pa.DataFrameSchema(
            columns=pa_columns,
            coerce=True,
        )

    def validate_dataframe(self, dataframe):
        try:
            self._schema(dataframe)
        except pa.errors.SchemaError as exc:
            raise InvalidDatasetError(
                f"An error occurred while validating the csv on column: '{exc.schema.name}'\n{exc.failure_cases}"
            )

    @property
    def schema(self):
        return self._schema

    def _pa_type_from_sql_type(self, sql_type):
        return {"text": pa.String, "int": pd.Int64Dtype(), "real": pa.Float}.get(
            sql_type
        )

    def _get_pa_checks(self, cdes_with_min_max, cdes_with_enumerations, column):
        checks = []
        if column in cdes_with_min_max:
            min, max = cdes_with_min_max[column]
            if max:
                checks.append(pa.Check(lambda s: s <= max))
            if min:
                checks.append(pa.Check(lambda s: s >= min))

        if column in cdes_with_enumerations:
            checks.append(
                pa.Check(lambda s: s.isin(cdes_with_enumerations[column] + ["None"]))
            )
        return checks



import os

import toml

CONFIG_PATH = "/home/config.toml"


def credentials_from_config():
    try:
        return toml.load(os.getenv("CONFIG_PATH", CONFIG_PATH))
    except FileNotFoundError:
        return {
            "DB_IP": "",
            "DB_PORT": "",
            "MONETDB_ADMIN_USERNAME": "",
            "MONETDB_LOCAL_USERNAME": "",
            "MONETDB_LOCAL_PASSWORD": "",
            "MONETDB_PUBLIC_USERNAME": "",
            "MONETDB_PUBLIC_PASSWORD": "",
            "DB_NAME": "",
            "SQLITE_DB_PATH": "",
        }



import sqlalchemy as sql

from mipdb.exceptions import UserInputError


class Schema:
    name: str
    schema: sql.MetaData

    def __init__(self, name) -> None:
        self.name = name
        self._validate_schema_name()
        self.schema = sql.MetaData(schema=self.name)

    def __repr__(self) -> str:
        return f"Schema(name={self.name})"

    def create(self, db):
        db.create_schema(self.name)

    def drop(self, db):
        db.drop_schema(self.name)

    def _validate_schema_name(self):
        if '"' in self.name:
            raise UserInputError(
                f"Data model's name: {self.name} contains prohibited character double quotes (\")"
            )



import pandas as pd


DATASET_COLUMN_NAME = "dataset"


class DataFrame:
    _data: pd.DataFrame
    _datasets: list

    def __init__(self, data: pd.DataFrame) -> None:
        # Pandas will insert nan values where there is an empty value in the csv.
        # In order to be able to insert the values through the sqlalchemy we need to replace nan with None.
        self._data = data.astype(object).where(pd.notnull(data), None)
        self._datasets = self._data[DATASET_COLUMN_NAME].unique()

    @property
    def data(self):
        return self._data

    @property
    def datasets(self):
        return self._datasets

    def to_dict(self):
        return self._data.to_dict("records")



import json

from mipdb.exceptions import UserInputError


class Properties:
    def __init__(self, properties) -> None:
        self.properties = properties
        if not self.properties:
            self.properties = {"tags": [], "properties": {}}

    def remove_tag(self, tag):
        if tag in self.properties["tags"]:
            self.properties["tags"].remove(tag)
        else:
            raise UserInputError("Tag does not exist")

    def add_tag(self, tag):
        if tag not in self.properties["tags"]:
            self.properties["tags"].append(tag)
        else:
            raise UserInputError("Tag already exists")

    def remove_property(self, key, value):
        if (key, value) in self.properties["properties"].items():
            self.properties["properties"].pop(key)
        else:
            raise UserInputError("Property does not exist")

    def add_property(self, key, value, force):
        if key in self.properties["properties"] and not force:
            raise UserInputError(
                "Property already exists.\n"
                "If you want to force override the property, please use the  '--force' flag"
            )
        else:
            self.properties["properties"][key] = value



import json
from dataclasses import dataclass

from mipdb.exceptions import InvalidDataModelError, UserInputError


@dataclass
class CommonDataElement:
    code: str
    metadata: str

    @classmethod
    def from_metadata(cls, metadata):
        code = metadata["code"]
        if not code.isidentifier():
            raise UserInputError(f"CDE: {code} is not a valid python identifier")

        validate_metadata(code, metadata)
        metadata = json.dumps(metadata)

        return cls(
            code,
            metadata,
        )

    def get_enumerations(self):
        metadata = json.loads(self.metadata)
        return metadata["enumerations"] if "enumerations" in metadata else []


def flatten_cdes(schema_data):
    cdes = []

    if "variables" in schema_data:
        for metadata in schema_data["variables"]:
            metadata = reformat_metadata(metadata)
            cdes += [CommonDataElement.from_metadata(metadata)]
    if "groups" in schema_data:
        cdes += [
            metadata
            for group_data in schema_data["groups"]
            for metadata in flatten_cdes(group_data)
        ]
    return cdes


def get_sql_type_per_column(cdes):
    return {code: json.loads(cde.metadata)["sql_type"] for code, cde in cdes.items()}


def get_cdes_with_min_max(cdes, columns):
    cdes_with_min_max = {}
    for code, cde in cdes.items():
        if code not in columns:
            continue
        metadata = json.loads(cde.metadata)
        max_value = metadata["max"] if "max" in metadata else None
        min_value = metadata["min"] if "min" in metadata else None
        if code in columns and min_value or max_value:
            cdes_with_min_max[code] = (min_value, max_value)
    return cdes_with_min_max


def get_cdes_with_enumerations(cdes, columns):
    return {
        code: [
            enum_code
            for enum_code, enum_label in json.loads(cde.metadata)[
                "enumerations"
            ].items()
        ]
        for code, cde in cdes.items()
        if json.loads(cde.metadata)["is_categorical"] and code in columns
    }


def get_dataset_enums(cdes):
    return json.loads(cdes["dataset"].metadata)["enumerations"]


def validate_dataset_present_on_cdes_with_proper_format(cdes):
    dataset_cde = [cde for cde in cdes if cde.code == "dataset"]
    if not dataset_cde:
        raise InvalidDataModelError("There is no 'dataset' CDE in the data model.")
    dataset_metadata = json.loads(dataset_cde[0].metadata)
    if not dataset_metadata["is_categorical"]:
        raise InvalidDataModelError(
            "CDE 'dataset' must have the 'isCategorical' property equal to 'true'."
        )
    if dataset_metadata["sql_type"] != "text":
        raise InvalidDataModelError(
            "CDE 'dataset' must have the 'sql_type' property equal to 'text'."
        )


def validate_longitudinal_data_model(cdes):
    subject_id_metadata = None
    visit_id_metadata = None

    for cde in cdes:
        if cde.code == "subjectid":
            subject_id_metadata = json.loads(cde.metadata)
        elif cde.code == "visitid":
            visit_id_metadata = json.loads(cde.metadata)

    if not subject_id_metadata:
        raise InvalidDataModelError(
            "There is no 'subjectid' CDE in the longitudinal data model."
        )
    if not visit_id_metadata:
        raise InvalidDataModelError(
            "There is no 'visitid' CDE in the longitudinal data model."
        )

    validate_visitid_cde(visit_id_metadata)


def validate_visitid_cde(metadata):
    if not metadata["is_categorical"]:
        raise InvalidDataModelError(
            "CDE 'visitid' must have the 'isCategorical' property equal to 'true'."
        )
    if metadata["sql_type"] != "text":
        raise InvalidDataModelError(
            "CDE 'visitid' must have the 'sql_type' property equal to 'text'."
        )
    if "enumerations" not in metadata:
        raise InvalidDataModelError(
            "CDE 'visitid' must contain the 'enumerations' property."
        )


def reformat_metadata(metadata):
    new_key_assign = {
        "isCategorical": "is_categorical",
        "minValue": "min",
        "maxValue": "max",
    }
    for old_key, new_key in new_key_assign.items():
        if old_key in metadata:
            metadata[new_key] = metadata.pop(old_key)

    if "enumerations" in metadata:
        metadata["enumerations"] = {
            enumeration["code"]: enumeration["label"]
            for enumeration in metadata["enumerations"]
        }
    return metadata


def validate_metadata(code, metadata):
    for element in ["is_categorical", "code", "sql_type", "label", "type"]:
        if element not in metadata:
            raise InvalidDataModelError(
                f"Element: {element} is missing from the CDE {code}"
            )
    if metadata["is_categorical"] and "enumerations" not in metadata:
        raise InvalidDataModelError(
            f"The CDE {code} has 'is_categorical' set to True but there are no enumerations."
        )
    if {"min", "max"} < set(metadata) and metadata["min"] >= metadata["max"]:
        raise InvalidDataModelError(f"The CDE {code} has min greater than the max.")

    valid_metadata_types = ["nominal", "real", "integer", "text"]

    if metadata["type"] not in valid_metadata_types:
        raise InvalidDataModelError(
            f"The CDE {code} has an 'type' the only valid types are:{valid_metadata_types} "
        )



from contextlib import contextmanager
from typing import Union

import sqlalchemy as sql
from sqlalchemy import text, inspect
from sqlalchemy.engine import Engine, Connection as SAConnection

from mipdb.exceptions import DataBaseError

PRIMARYDATA_TABLE = "primary_data"


def handle_errors(func):
    """Decorator to handle DB exceptions and raise DataBaseError."""

    @contextmanager
    def _handle_errors():
        try:
            yield
        except (sql.exc.OperationalError, sql.exc.IntegrityError) as exc:
            _, msg = exc.orig.args[0].split("!")
            raise DataBaseError(msg)

    def wrapper(*args, **kwargs):
        with _handle_errors():
            return func(*args, **kwargs)

    return wrapper


class DBExecutor:
    """Class to handle SQL execution using SQLAlchemy's Engine and Connection."""

    def __init__(self, executor: Union[Engine, SAConnection]):
        self._executor = executor

    def _get_connection(self):
        # SQLAlchemy 2.x: if executor is Engine, open a new Connection; if already Connection, use it
        if isinstance(self._executor, Engine):
            return self._executor.connect()
        if isinstance(self._executor, SAConnection):
            return self._executor
        raise DataBaseError(f"Unsupported executor type: {type(self._executor)}")

    @handle_errors
    def execute(self, query, *args, **kwargs) -> sql.engine.CursorResult | None:
        conn = self._get_connection()
        own_conn = isinstance(self._executor, sql.engine.Engine)
        try:
            stmt = text(query) if isinstance(query, str) else query

            params = kwargs or None
            result = conn.execute(stmt, params) if params else conn.execute(stmt, *args)

            if own_conn:
                conn.commit()
            return result
        finally:
            if own_conn:
                conn.close()

    def create_schema(self, schema_name):
        self.execute(sql.schema.CreateSchema(schema_name))

    def drop_schema(self, schema_name):
        self.execute(f'DROP SCHEMA "{schema_name}" CASCADE')

    def get_schemas(self):
        res = self.execute("SELECT name FROM sys.schemas WHERE system=FALSE")
        return [schema for schema, *_ in res]

    def get_data_count_by_dataset(self, schema_fullname):
        res = self.execute(
            f"""
            SELECT dataset, COUNT(dataset) as count
            FROM "{schema_fullname}"."{PRIMARYDATA_TABLE}"
            GROUP BY dataset
        """
        )
        return list(res)

    def get_row_count(self, table):
        res = self.execute(f"SELECT COUNT(*) FROM {table}")
        return res.scalar_one() if res is not None else 0

    def get_column_distinct(self, column, table):
        datasets = self.execute(f"SELECT DISTINCT({column}) FROM {table};")
        return [dataset[0] for dataset in datasets]

    def table_exists(self, table) -> bool:
        inspector = inspect(self._executor)
        schema = table.schema or None
        return inspector.has_table(table.name, schema)

    @handle_errors
    def create_table(self, table):
        table.create(bind=self._executor)

    @handle_errors
    def grant_select_access_rights(self, table, user):
        fullname = (
            f'"{table.schema}"."{table.name}"' if table.schema else f'"{table.name}"'
        )
        self.execute(f"GRANT SELECT ON TABLE {fullname} TO {user} WITH GRANT OPTION;")

    def copy_csv_in_table(self, file_location, records, offset, table_name):
        records_query = f"{records} RECORDS" if records else ""
        self.execute(
            f"""
            COPY {records_query} OFFSET {offset} INTO {table_name}
            FROM '{file_location}'
            USING DELIMITERS ',', E'\n', '"'
            NULL AS '';
        """
        )

    def copy_data_table_to_another_table(self, copy_into_table, copy_from_table):
        table_columns = [
            col.name for col in copy_into_table.table.columns if col.name != "row_id"
        ]
        csv_columns = [col.name for col in copy_from_table.table.columns]
        select_columns = [
            f'"{col}"' if col in csv_columns else "NULL" for col in table_columns
        ]
        self.execute(
            f"""
            INSERT INTO "{copy_into_table.table.schema}".{copy_into_table.table.name} ({','.join(f'"{col}"' for col in table_columns)})
            SELECT {', '.join(select_columns)}
            FROM {copy_from_table.table.name};
        """
        )

    @handle_errors
    def drop_table(self, table):
        table.drop(bind=self._executor)

    def delete_table_values(self, table):
        self.execute(table.delete())

    def insert_values_to_table(self, table, values):
        self.execute(table.insert(), values)

    def get_current_user(self):
        res = self.execute("SELECT CURRENT_USER")
        return res[0][0] if res else None

    def get_executor(self):
        return self._executor


class MonetDBConnection(DBExecutor):
    """Concrete connection object for MonetDB within transaction boundaries."""

    def __init__(self, conn: SAConnection) -> None:
        super().__init__(conn)


class MonetDB(DBExecutor):
    """Concrete DataBase object for connecting to a MonetDB instance."""

    def __init__(self, url: str, echo=False) -> None:
        super().__init__(sql.create_engine(url, echo=echo))

    @classmethod
    def from_config(cls, dbconfig) -> "MonetDB":
        username, password, ip, port, dbfarm = (
            dbconfig["username"],
            dbconfig["password"],
            dbconfig["ip"],
            dbconfig["port"],
            dbconfig["dbfarm"],
        )
        url = f"monetdb://{username}:{password}@{ip}:{port}/{dbfarm}"
        return cls(url)

    @contextmanager
    def begin(self) -> MonetDBConnection:
        with self._executor.begin() as conn:
            yield MonetDBConnection(conn)



from typing import List, Any, Dict
from enum import Enum

import sqlalchemy as sql
from sqlalchemy import MetaData, ForeignKey, inspect, text, select, func
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.exc import MultipleResultsFound

from mipdb.exceptions import DataBaseError

METADATA_TABLE = "variables_metadata"
PRIMARYDATA_TABLE = "primary_data"

metadata = MetaData()
Base = declarative_base(metadata=metadata)


class Status:
    ENABLED = "ENABLED"
    DISABLED = "DISABLED"


class DBType(Enum):
    monetdb = "monetdb"
    sqlite = "sqlite"


class SQLTYPES:
    INTEGER = sql.Integer
    STRING = sql.String(255)
    FLOAT = sql.Float
    JSON = sql.JSON


class DataModel(Base):
    __tablename__ = "data_models"
    data_model_id = sql.Column(sql.Integer, primary_key=True, autoincrement=True)
    code = sql.Column(sql.String, nullable=False)
    version = sql.Column(sql.String, nullable=False)
    label = sql.Column(sql.String, nullable=False)
    status = sql.Column(sql.String, nullable=False)
    properties = sql.Column(sql.JSON, nullable=True)


class Dataset(Base):
    __tablename__ = "datasets"
    dataset_id = sql.Column(sql.Integer, primary_key=True, autoincrement=True)
    data_model_id = sql.Column(
        sql.Integer,
        ForeignKey("data_models.data_model_id"),
        nullable=False,
    )
    code = sql.Column(sql.String, nullable=False)
    label = sql.Column(sql.String, nullable=False)
    status = sql.Column(sql.String, nullable=False)
    csv_path = sql.Column(sql.String, nullable=True)
    properties = sql.Column(sql.JSON, nullable=True)


class SQLiteDB:
    """Class representing a SQLite database interface."""

    def __init__(self, url: str, echo=False) -> None:
        self._executor = sql.create_engine(url, echo=echo)
        self.Session = sessionmaker(bind=self._executor, future=True)

    @classmethod
    def from_config(cls, dbconfig: Dict[str, Any]) -> "SQLiteDB":
        db_path = dbconfig["db_path"]
        url = f"sqlite:///{db_path}"
        return SQLiteDB(url)

    def execute(self, query: str, *args, **kwargs) -> List[Any]:
        """Execute a statement without returning rows."""
        with self._executor.connect() as conn:
            conn.execute(text(query), *args, **kwargs)
            conn.commit()
        return []

    def execute_fetchall(self, query: str, *args, **kwargs) -> List[Any]:
        """Execute a query and return all rows."""
        with self._executor.connect() as conn:
            result = conn.execute(text(query), *args, **kwargs)
            return result.fetchall()

    def insert_values_to_table(self, table: sql.Table, values: List[Dict[str, Any]]) -> None:
        session = self.Session()
        try:
            session.execute(table.insert(), values)
            session.commit()
        finally:
            session.close()

    def get_data_model_status(self, data_model_id: int) -> Any:
        session = self.Session()
        try:
            result = session.execute(
                select(Base.metadata.tables["data_models"].c.status)
                .where(Base.metadata.tables["data_models"].c.data_model_id == data_model_id)
            ).scalar_one_or_none()
        finally:
            session.close()
        return result

    def update_data_model_status(self, status: str, data_model_id: int) -> None:
        session = self.Session()
        try:
            session.execute(
                sql.update(Base.metadata.tables["data_models"])
                .where(Base.metadata.tables["data_models"].c.data_model_id == data_model_id)
                .values(status=status)
            )
            session.commit()
        finally:
            session.close()

    def get_dataset_status(self, dataset_id: int) -> Any:
        session = self.Session()
        try:
            result = session.execute(
                select(Base.metadata.tables["datasets"].c.status)
                .where(Base.metadata.tables["datasets"].c.dataset_id == dataset_id)
            ).scalar_one_or_none()
        finally:
            session.close()
        return result

    def get_metadata(self, data_model: str) -> Dict[str, Any]:
        table = sql.Table(
            f"{data_model}_{METADATA_TABLE}",
            Base.metadata,
            sql.Column("code", SQLTYPES.STRING, primary_key=True),
            sql.Column("metadata", SQLTYPES.JSON),
            extend_existing=True,
        )
        session = self.Session()
        try:
            result = session.execute(select(table.c.code, table.c.metadata)).all()
        finally:
            session.close()
        return {code: meta for code, meta in result}

    def update_dataset_status(self, status: str, dataset_id: int) -> None:
        session = self.Session()
        try:
            session.execute(
                sql.update(Base.metadata.tables["datasets"])
                .where(Base.metadata.tables["datasets"].c.dataset_id == dataset_id)
                .values(status=status)
            )
            session.commit()
        finally:
            session.close()

    def get_dataset(self, dataset_id: int, columns: List[str]) -> Any:
        session = self.Session()
        try:
            cols = [getattr(Base.metadata.tables["datasets"].c, col) for col in columns]
            result = session.execute(
                select(*cols).where(Base.metadata.tables["datasets"].c.dataset_id == dataset_id)
            ).one_or_none()
        finally:
            session.close()
        return result

    def get_data_model(self, data_model_id: int, columns: List[str]) -> Any:
        session = self.Session()
        try:
            cols = [getattr(Base.metadata.tables["data_models"].c, col) for col in columns]
            result = session.execute(
                select(*cols).where(Base.metadata.tables["data_models"].c.data_model_id == data_model_id)
            ).one_or_none()
        finally:
            session.close()
        return result

    from sqlalchemy import select

    # ...

    def get_values(
            self,
            table: sql.Table,
            columns: List[str] | None = None,
            where_conditions: Dict[str, Any] | None = None,
    ) -> List[sql.Row]:
        """Return rows (SQLAlchemy Row objects) respecting an optional WHERE."""
        stmt = select(
            *[table.c[col] for col in (columns or [c.name for c in table.columns])]
        )
        if where_conditions:
            for col, val in where_conditions.items():
                stmt = stmt.where(table.c[col] == val)

        with self.Session() as session:
            return session.execute(stmt).all()  # <-- rows, not dicts

    def get_data_models(self, columns: List[str]) -> List[Dict[str, Any]]:
        session = self.Session()
        try:
            cols = [getattr(Base.metadata.tables["data_models"].c, col) for col in columns]
            rows = session.execute(select(*cols)).all()
        finally:
            session.close()
        return [dict(zip(columns, row)) for row in rows]

    def get_dataset_count_by_data_model_id(self) -> List[Dict[str, Any]]:
        session = self.Session()
        try:
            stmt = (
                select(
                    Base.metadata.tables["datasets"].c.data_model_id,
                    func.count(Base.metadata.tables["datasets"].c.data_model_id).label("count"),
                )
                .group_by(Base.metadata.tables["datasets"].c.data_model_id)
            )
            rows = session.execute(stmt).all()
        finally:
            session.close()
        return [dict(data_model_id=row[0], count=row[1]) for row in rows]

    def get_row_count(self, table_name: str) -> int:
        session = self.Session()
        try:
            count = session.execute(select(func.count()).select_from(text(table_name))).scalar_one()
        finally:
            session.close()
        return count

    def drop_table(self, table_name: str) -> None:
        session = self.Session()
        try:
            table = sql.Table(table_name, metadata, autoload_with=self._executor)
            table.drop(bind=self._executor)
            session.commit()
        finally:
            session.close()

    def delete_from(self, table: sql.Table, where_conditions: Dict[str, Any]) -> None:
        session = self.Session()
        try:
            stmt = table.delete()
            if where_conditions:
                for col_name, val in where_conditions.items():
                    stmt = stmt.where(table.c[col_name] == val)
            session.execute(stmt)
            session.commit()
        finally:
            session.close()

    def get_dataset_properties(self, dataset_id: int) -> Any:
        session = self.Session()
        try:
            result = session.execute(
                select(Base.metadata.tables["datasets"].c.properties)
                .where(Base.metadata.tables["datasets"].c.dataset_id == dataset_id)
            ).scalar_one_or_none()
        finally:
            session.close()
        return result or {}

    def get_data_model_properties(self, data_model_id: int) -> Any:
        session = self.Session()
        try:
            result = session.execute(
                select(Base.metadata.tables["data_models"].c.properties)
                .where(Base.metadata.tables["data_models"].c.data_model_id == data_model_id)
            ).scalar_one_or_none()
        finally:
            session.close()
        return result or {}

    def set_data_model_properties(self, properties: Dict[str, Any], data_model_id: int) -> None:
        session = self.Session()
        try:
            session.execute(
                sql.update(Base.metadata.tables["data_models"])
                .where(Base.metadata.tables["data_models"].c.data_model_id == data_model_id)
                .values(properties=properties)
            )
            session.commit()
        finally:
            session.close()

    def set_dataset_properties(self, properties: Dict[str, Any], dataset_id: int) -> None:
        session = self.Session()
        try:
            session.execute(
                sql.update(Base.metadata.tables["datasets"])
                .where(Base.metadata.tables["datasets"].c.dataset_id == dataset_id)
                .values(properties=properties)
            )
            session.commit()
        finally:
            session.close()

    def get_data_model_id(self, code: str, version: str) -> int:
        session = self.Session()
        try:
            stmt = (
                select(Base.metadata.tables["data_models"].c.data_model_id)
                .where(
                    Base.metadata.tables["data_models"].c.code == code,
                    Base.metadata.tables["data_models"].c.version == version,
                )
            )
            data_model_id = session.execute(stmt).scalar_one_or_none()
        except MultipleResultsFound:
            raise DataBaseError(f"Got more than one data_model ids for code={code} and version={version}.")
        finally:
            session.close()
        if not data_model_id:
            raise DataBaseError(f"Data_models table doesn't have a record with code={code}, version={version}")
        return data_model_id

    def get_max_data_model_id(self) -> int:
        session = self.Session()
        try:
            max_id = session.execute(
                select(func.max(Base.metadata.tables["data_models"].c.data_model_id))
            ).scalar_one()
        finally:
            session.close()
        return max_id

    def get_max_dataset_id(self) -> int:
        session = self.Session()
        try:
            max_id = session.execute(
                select(func.max(Base.metadata.tables["datasets"].c.dataset_id))
            ).scalar_one()
        finally:
            session.close()
        return max_id

    def get_dataset_id(self, code: str, data_model_id: int) -> int:
        session = self.Session()
        try:
            stmt = (
                select(Base.metadata.tables["datasets"].c.dataset_id)
                .where(
                    Base.metadata.tables["datasets"].c.code == code,
                    Base.metadata.tables["datasets"].c.data_model_id == data_model_id,
                )
            )
            dataset_id = session.execute(stmt).scalar_one_or_none()
        except MultipleResultsFound:
            raise DataBaseError(f"Got more than one dataset ids for code={code} and data_model_id={data_model_id}.")
        finally:
            session.close()
        if not dataset_id:
            raise DataBaseError(f"Datasets table doesn't have a record with code={code}, data_model_id={data_model_id}")
        return dataset_id

    def table_exists(self, table: sql.Table) -> bool:
        inspector = inspect(self._executor)
        schema = table.schema or None
        return inspector.has_table(table.name, schema)

    def create_table(self, table: sql.Table) -> None:
        table.create(bind=self._executor)

    def get_all_tables(self) -> List[str]:
        inspector = inspect(self._executor)
        return inspector.get_table_names()



import copy
import os
from abc import ABC, abstractmethod

import pandas as pd

from mipdb.databases.monetdb import MonetDB
from mipdb.data_frame_schema import DataFrameSchema
from mipdb.exceptions import ForeignKeyError, InvalidDatasetError, UserInputError
from mipdb.databases.monetdb_tables import (
    PrimaryDataTable,
    TemporaryTable,
    RECORDS_PER_COPY,
)
from mipdb.properties import Properties
from mipdb.reader import CSVDataFrameReader
from mipdb.dataelements import (
    flatten_cdes,
    validate_dataset_present_on_cdes_with_proper_format,
    validate_longitudinal_data_model,
    get_sql_type_per_column,
    get_cdes_with_min_max,
    get_cdes_with_enumerations,
    get_dataset_enums,
)
from mipdb.schema import Schema
from mipdb.databases.sqlite import SQLiteDB
from mipdb.databases.sqlite_tables import DataModelTable, DatasetsTable, MetadataTable
from mipdb.data_frame import DataFrame, DATASET_COLUMN_NAME

LONGITUDINAL = "longitudinal"


class UseCase(ABC):
    """Abstract use case class."""

    @abstractmethod
    def execute(self, *args, **kwargs) -> None:
        """Executes use case logic with arguments from CLI command."""


def is_db_initialized(db: SQLiteDB):
    if not (DataModelTable().exists(db) and DatasetsTable().exists(db)):
        raise UserInputError("You need to initialize the database!\nTry mipdb init")


class InitDB(UseCase):
    def __init__(self, db: SQLiteDB) -> None:
        self.db = db

    def execute(self) -> None:
        data_model_table, datasets_table = DataModelTable(), DatasetsTable()
        if not data_model_table.exists(self.db):
            data_model_table.create(self.db)
        if not datasets_table.exists(self.db):
            datasets_table.create(self.db)


class AddDataModel(UseCase):
    def __init__(self, sqlite_db: SQLiteDB, monetdb: MonetDB) -> None:
        self.sqlite_db = sqlite_db
        self.monetdb = monetdb
        is_db_initialized(sqlite_db)

    def execute(self, data_model_metadata) -> None:
        code, version = data_model_metadata["code"], data_model_metadata["version"]
        data_model = get_data_model_fullname(code, version)
        cdes = flatten_cdes(copy.deepcopy(data_model_metadata))
        data_model_table = DataModelTable()
        data_model_id = data_model_table.get_next_data_model_id(self.sqlite_db)
        self._create_primary_data_table(data_model, cdes)
        self._create_metadata_table(data_model, cdes)
        properties = Properties(
            data_model_table.get_data_model_properties(data_model_id, self.sqlite_db)
        )
        properties.add_property("cdes", data_model_metadata, True)
        values = dict(
            data_model_id=data_model_id,
            code=code,
            version=version,
            label=data_model_metadata["label"],
            status="ENABLED",
            properties=properties.properties,
        )
        data_model_table.insert_values(values, self.sqlite_db)
        self._tag_longitudinal_if_needed(data_model_metadata, code, version)

    def _create_primary_data_table(self, data_model, cdes):
        with self.monetdb.begin() as conn:
            schema = Schema(data_model)
            schema.create(conn)
            PrimaryDataTable.from_cdes(schema, cdes).create(conn)

    def _create_metadata_table(self, data_model, cdes):
        metadata_table = MetadataTable(data_model)
        metadata_table.create(self.sqlite_db)
        values = metadata_table.get_values_from_cdes(cdes)
        metadata_table.insert_values(values, self.sqlite_db)

    def _tag_longitudinal_if_needed(self, data_model_metadata, code, version):
        if LONGITUDINAL in data_model_metadata:
            longitudinal = data_model_metadata[LONGITUDINAL]
            if not isinstance(longitudinal, bool):
                raise UserInputError(
                    f"Longitudinal flag should be boolean, value given: {longitudinal}"
                )
            if longitudinal:
                TagDataModel(self.sqlite_db).execute(
                    code=code, version=version, tag=LONGITUDINAL
                )


class ValidateDataModel(UseCase):
    def execute(self, data_model_metadata) -> None:
        if "version" not in data_model_metadata:
            raise UserInputError(
                "You need to include a version on the CDEsMetadata.json"
            )
        cdes = flatten_cdes(copy.deepcopy(data_model_metadata))
        validate_dataset_present_on_cdes_with_proper_format(cdes)
        if LONGITUDINAL in data_model_metadata:
            longitudinal = data_model_metadata[LONGITUDINAL]
            if not isinstance(longitudinal, bool):
                raise UserInputError(
                    f"Longitudinal flag should be boolean, value given: {longitudinal}"
                )
            if longitudinal:
                validate_longitudinal_data_model(cdes)


class DeleteDataModel(UseCase):
    def __init__(self, sqlite_db: SQLiteDB, monetdb: MonetDB) -> None:
        self.sqlite_db = sqlite_db
        self.monetdb = monetdb
        is_db_initialized(sqlite_db)

    def execute(self, code, version, force) -> None:
        name = get_data_model_fullname(code, version)
        schema = Schema(name)
        data_model_table = DataModelTable()
        data_model_id = data_model_table.get_data_model_id(
            code, version, self.sqlite_db
        )
        if not force:
            self._validate_data_model_deletion(name, data_model_id)
        MetadataTable(data_model=name).drop(self.sqlite_db)
        self._delete_datasets(data_model_id, code, version)
        with self.monetdb.begin() as conn:
            schema.drop(conn)
        data_model_table.delete_data_model(code, version, self.sqlite_db)

    def _validate_data_model_deletion(self, data_model_name, data_model_id):
        datasets = DatasetsTable().get_dataset_codes(
            db=self.sqlite_db, columns=["code"], data_model_id=data_model_id
        )
        if datasets:
            raise ForeignKeyError(
                f"The Data Model:{data_model_name} cannot be deleted because it contains Datasets: {datasets}\nIf you want to force delete everything, please use the '--force' flag"
            )

    def _delete_datasets(self, data_model_id, data_model_code, data_model_version):
        datasets_table = DatasetsTable()
        dataset_codes = datasets_table.get_dataset_codes(
            data_model_id=data_model_id, columns=["code"], db=self.sqlite_db
        )
        for dataset_code in dataset_codes:
            DeleteDataset(sqlite_db=self.sqlite_db, monetdb=self.monetdb).execute(
                dataset_code,
                data_model_code=data_model_code,
                data_model_version=data_model_version,
            )


class ImportCSV(UseCase):
    def __init__(self, sqlite_db: SQLiteDB, monetdb: MonetDB) -> None:
        self.sqlite_db = sqlite_db
        self.monetdb = monetdb
        is_db_initialized(sqlite_db)

    def execute(
        self, csv_path, copy_from_file, data_model_code, data_model_version
    ) -> None:
        data_model_name = get_data_model_fullname(
            code=data_model_code, version=data_model_version
        )
        data_model = Schema(data_model_name)
        data_model_id = DataModelTable().get_data_model_id(
            data_model_code, data_model_version, self.sqlite_db
        )
        metadata_table = MetadataTable.from_db(data_model_name, self.sqlite_db)
        cdes = metadata_table.table
        dataset_enumerations = get_dataset_enums(cdes)
        sql_type_per_column = get_sql_type_per_column(cdes)

        with self.monetdb.begin() as monetdb_conn:
            imported_datasets = (
                self.import_csv_with_volume(
                    csv_path, sql_type_per_column, data_model, monetdb_conn
                )
                if copy_from_file
                else self._import_csv(csv_path, data_model, monetdb_conn)
            )

        existing_datasets = DatasetsTable().get_dataset_codes(
            columns=["code"], data_model_id=data_model_id, db=self.sqlite_db
        )
        dataset_id = self._get_next_dataset_id()
        for dataset in set(imported_datasets) - set(existing_datasets):
            values = dict(
                data_model_id=data_model_id,
                dataset_id=dataset_id,
                code=dataset,
                label=dataset_enumerations[dataset],
                csv_path=csv_path if copy_from_file else None,
                status="ENABLED",
                properties=None,
            )
            DatasetsTable().insert_values(values, self.sqlite_db)
            dataset_id += 1

    def _get_next_dataset_id(self):
        return DatasetsTable().get_next_dataset_id(self.sqlite_db)

    def _create_temporary_table(self, dataframe_sql_type_per_column, db):
        temporary_table = TemporaryTable(dataframe_sql_type_per_column, db)
        temporary_table.create(db)
        return temporary_table

    def import_csv_with_volume(self, csv_path, sql_type_per_column, data_model, conn):
        csv_columns = pd.read_csv(csv_path, nrows=0).columns.tolist()
        dataframe_sql_type_per_column = {
            dataframe_column: sql_type_per_column[dataframe_column]
            for dataframe_column in csv_columns
        }
        temporary_table = self._create_temporary_table(
            dataframe_sql_type_per_column, conn
        )
        imported_datasets = self.insert_csv_to_db(
            csv_path, temporary_table, data_model, conn
        )
        temporary_table.drop(conn)
        return imported_datasets

    # The monetdb copy into prohibites the importion of a csv,
    # to a table if the csv contains fewer columns than the table.
    # In our case we need to load the csv a table named 'primary_data'.
    # Minimal example:
    # TABLE 'primary_data': col1 col2 | CSV: col1
    #                       1    2    |      5
    #                       2    4    |      6
    #                       3    6    |      7
    #                       4    8    |      8
    # We need to create a 'temp' table which will mirror the columns of the csv.
    # Thus making the copy into a valid choice again.
    # The workaround for that is that we first copy the value into the 'temp'.
    # Once the data is loaded in the 'temp',
    # we can now pass the values of the 'temp' to the 'primary_data' table using a query with the format:
    # INSERT INTO table1 (columns)
    # SELECT columns
    # FROM table2;
    # What about the missing columns at that the 'primary_data' contains but the 'temp' does not contain,
    # we will simply add null in place of the missing columns.
    # In our case the query will have the form:
    # INSERT INTO 'primary_data' (col1, col2)
    # SELECT col1, NULL
    # FROM 'temp'

    def insert_csv_to_db(self, csv_path, temporary_table, data_model, db):
        primary_data_table = PrimaryDataTable.from_db(data_model, db)
        offset, imported_datasets = 2, []
        while True:
            temporary_table.load_csv(
                csv_path=csv_path, offset=offset, records=RECORDS_PER_COPY, db=db
            )
            offset += RECORDS_PER_COPY
            table_count = temporary_table.get_row_count(db=db)
            if not table_count:
                break
            imported_datasets = set(imported_datasets) | set(
                temporary_table.get_column_distinct(DATASET_COLUMN_NAME, db)
            )
            db.copy_data_table_to_another_table(primary_data_table, temporary_table)
            temporary_table.delete(db)
            if table_count < RECORDS_PER_COPY:
                break
        return imported_datasets

    def _import_csv(self, csv_path, data_model, conn):
        imported_datasets, primary_data_table = [], PrimaryDataTable.from_db(
            data_model, conn
        )
        with CSVDataFrameReader(csv_path).get_reader() as reader:
            for dataset_data in reader:
                dataframe = DataFrame(dataset_data)
                imported_datasets = set(imported_datasets) | set(dataframe.datasets)
                primary_data_table.insert_values(dataframe.to_dict(), conn)
        return imported_datasets


def are_data_valid_longitudinal(csv_path):
    df = pd.read_csv(csv_path, usecols=["subjectid", "visitid"])
    check_unique_longitudinal_dataset_primary_keys(df)
    check_subjectid_is_full(df)
    check_visitid_is_full(df)


def check_unique_longitudinal_dataset_primary_keys(df):
    duplicates = df[df.duplicated(subset=["visitid", "subjectid"], keep=False)]
    if not duplicates.empty:
        raise InvalidDatasetError(
            f"Invalid csv: the following visitid and subjectid pairs are duplicated:\n{duplicates}"
        )


def check_subjectid_is_full(df):
    if df["subjectid"].isnull().any():
        raise InvalidDatasetError("Column 'subjectid' should never contain null values")


def check_visitid_is_full(df):
    if df["visitid"].isnull().any():
        raise InvalidDatasetError("Column 'visitid' should never contain null values")


class ValidateDataset(UseCase):
    """
    We separate the data validation from the importation to make sure that a csv is valid as a whole before committing it to the main table.
    In the data validation we use chunking in order to reduce the memory footprint of the process.
    Database constraints must NOT be used as part of the validation process since that could result in partially imported csvs.
    """

    def __init__(self, sqlite_db: SQLiteDB, monetdb: MonetDB) -> None:
        self.sqlite_db = sqlite_db
        self.monetdb = monetdb
        is_db_initialized(sqlite_db)

    def execute(
        self, csv_path, copy_from_file, data_model_code, data_model_version
    ) -> None:
        data_model = get_data_model_fullname(
            code=data_model_code, version=data_model_version
        )
        csv_columns = pd.read_csv(csv_path, nrows=0).columns.tolist()
        if DATASET_COLUMN_NAME not in csv_columns:
            raise InvalidDatasetError(
                "The 'dataset' column is required to exist in the csv."
            )
        metadata_table = MetadataTable.from_db(data_model, self.sqlite_db)
        cdes = metadata_table.table
        sql_type_per_column = get_sql_type_per_column(cdes)
        cdes_with_min_max = get_cdes_with_min_max(cdes, csv_columns)
        cdes_with_enumerations = get_cdes_with_enumerations(cdes, csv_columns)
        dataset_enumerations = get_dataset_enums(cdes)
        if self.is_data_model_longitudinal(data_model_code, data_model_version):
            are_data_valid_longitudinal(csv_path)

        if copy_from_file:
            with self.monetdb.begin() as monetdb_conn:
                validated_datasets = self.validate_csv_with_volume(
                    csv_path,
                    sql_type_per_column,
                    cdes_with_min_max,
                    cdes_with_enumerations,
                    monetdb_conn,
                )
        else:
            validated_datasets = self.validate_csv(
                csv_path, sql_type_per_column, cdes_with_min_max, cdes_with_enumerations
            )
        self.verify_datasets_exist_in_enumerations(
            validated_datasets, dataset_enumerations
        )

    def is_data_model_longitudinal(self, data_model_code, data_model_version):
        data_model_id = DataModelTable().get_data_model_id(
            data_model_code, data_model_version, self.sqlite_db
        )
        properties = DataModelTable().get_data_model_properties(
            data_model_id, self.sqlite_db
        )
        return LONGITUDINAL in properties["tags"]

    def validate_csv(
        self, csv_path, sql_type_per_column, cdes_with_min_max, cdes_with_enumerations
    ):
        imported_datasets = []
        csv_columns = pd.read_csv(csv_path, nrows=0).columns.tolist()
        dataframe_schema = DataFrameSchema(
            sql_type_per_column, cdes_with_min_max, cdes_with_enumerations, csv_columns
        )
        with CSVDataFrameReader(csv_path).get_reader() as reader:
            for dataset_data in reader:
                dataframe = DataFrame(dataset_data)
                dataframe_schema.validate_dataframe(dataframe.data)
                imported_datasets = set(imported_datasets) | set(dataframe.datasets)
        return imported_datasets

    def validate_csv_with_volume(
        self,
        csv_path,
        sql_type_per_column,
        cdes_with_min_max,
        cdes_with_enumerations,
        conn,
    ):
        csv_columns = pd.read_csv(csv_path, nrows=0).columns.tolist()
        dataframe_sql_type_per_column = self._get_dataframe_sql_type_per_column(
            csv_columns, sql_type_per_column
        )
        temporary_table = self._create_temporary_table(
            dataframe_sql_type_per_column, conn
        )
        validated_datasets = temporary_table.validate_csv(
            csv_path, cdes_with_min_max, cdes_with_enumerations, conn
        )
        temporary_table.drop(conn)
        return validated_datasets

    def _get_dataframe_sql_type_per_column(self, csv_columns, sql_type_per_column):
        if set(csv_columns) <= set(sql_type_per_column.keys()):
            return {
                dataframe_column: sql_type_per_column[dataframe_column]
                for dataframe_column in csv_columns
            }
        raise InvalidDatasetError(
            f"Columns:{set(csv_columns) - set(sql_type_per_column.keys()) - {'row_id'}} are not present in the CDEs"
        )

    def _create_temporary_table(self, dataframe_sql_type_per_column, conn):
        temporary_table = TemporaryTable(dataframe_sql_type_per_column, conn)
        temporary_table.create(conn)
        return temporary_table

    def verify_datasets_exist_in_enumerations(self, datasets, dataset_enumerations):
        non_existing_datasets = [
            dataset for dataset in datasets if dataset not in dataset_enumerations
        ]
        if non_existing_datasets:
            raise InvalidDatasetError(
                f"The values:'{non_existing_datasets}' are not present in the enumerations of the CDE 'dataset'."
            )


class ValidateDatasetNoDatabase(UseCase):
    """
    We separate the data validation from the importation to make sure that a csv is valid as a whole before committing it to the main table.
    In the data validation we use chunking in order to reduce the memory footprint of the process.
    Database constraints must NOT be used as part of the validation process since that could result in partially imported csvs.
    """

    def execute(self, csv_path, data_model_metadata) -> None:

        csv_columns = pd.read_csv(csv_path, nrows=0).columns.tolist()
        if DATASET_COLUMN_NAME not in csv_columns:
            raise InvalidDatasetError(
                "The 'dataset' column is required to exist in the csv."
            )
        cdes = flatten_cdes(copy.deepcopy(data_model_metadata))
        cdes = {cde.code: cde for cde in cdes}
        sql_type_per_column = get_sql_type_per_column(cdes)
        cdes_with_min_max = get_cdes_with_min_max(cdes, csv_columns)
        cdes_with_enumerations = get_cdes_with_enumerations(cdes, csv_columns)
        dataset_enumerations = get_dataset_enums(cdes)
        if LONGITUDINAL in data_model_metadata:
            longitudinal = data_model_metadata[LONGITUDINAL]
            if not isinstance(longitudinal, bool):
                raise UserInputError(
                    f"Longitudinal flag should be boolean, value given: {longitudinal}"
                )
            if longitudinal:
                are_data_valid_longitudinal(csv_path)
        validated_datasets = self.validate_csv(
            csv_path,
            sql_type_per_column,
            cdes_with_min_max,
            cdes_with_enumerations,
        )
        self.verify_datasets_exist_in_enumerations(
            datasets=validated_datasets,
            dataset_enumerations=dataset_enumerations,
        )

    def validate_csv(
        self, csv_path, sql_type_per_column, cdes_with_min_max, cdes_with_enumerations
    ):
        imported_datasets = []

        csv_columns = pd.read_csv(csv_path, nrows=0).columns.tolist()
        dataframe_schema = DataFrameSchema(
            sql_type_per_column, cdes_with_min_max, cdes_with_enumerations, csv_columns
        )
        with CSVDataFrameReader(csv_path).get_reader() as reader:
            for dataset_data in reader:
                dataframe = DataFrame(dataset_data)
                dataframe_schema.validate_dataframe(dataframe.data)
                imported_datasets = set(imported_datasets) | set(dataframe.datasets)
        return imported_datasets

    def verify_datasets_exist_in_enumerations(self, datasets, dataset_enumerations):
        non_existing_datasets = [
            dataset for dataset in datasets if dataset not in dataset_enumerations
        ]
        if non_existing_datasets:
            raise InvalidDatasetError(
                f"The values:'{non_existing_datasets}' are not present in the enumerations of the CDE 'dataset'."
            )


class DeleteDataset(UseCase):
    def __init__(self, sqlite_db: SQLiteDB, monetdb: MonetDB) -> None:
        self.sqlite_db = sqlite_db
        self.monetdb = monetdb
        is_db_initialized(sqlite_db)

    def execute(self, dataset_code, data_model_code, data_model_version) -> None:
        data_model_fullname = get_data_model_fullname(
            code=data_model_code, version=data_model_version
        )
        with self.monetdb.begin() as conn:
            primary_data_table = PrimaryDataTable.from_db(
                Schema(data_model_fullname), conn
            )
            primary_data_table.remove_dataset(dataset_code, data_model_fullname, conn)
        data_model_id = DataModelTable().get_data_model_id(
            data_model_code, data_model_version, self.sqlite_db
        )
        dataset_id = DatasetsTable().get_dataset_id(
            dataset_code, data_model_id, self.sqlite_db
        )
        DatasetsTable().delete_dataset(dataset_id, data_model_id, self.sqlite_db)


class EnableDataModel(UseCase):
    def __init__(self, db: SQLiteDB) -> None:
        self.db = db

    def execute(self, code, version) -> None:
        data_model_table = DataModelTable()
        data_model_id = data_model_table.get_data_model_id(code, version, self.db)
        current_status = data_model_table.get_data_model_status(data_model_id, self.db)
        if current_status != "ENABLED":
            data_model_table.set_data_model_status("ENABLED", data_model_id, self.db)
        else:
            raise UserInputError("The data model was already enabled")


class DisableDataModel(UseCase):
    def __init__(self, db: SQLiteDB) -> None:
        self.db = db

    def execute(self, code, version) -> None:
        data_model_table = DataModelTable()
        data_model_id = data_model_table.get_data_model_id(code, version, self.db)
        current_status = data_model_table.get_data_model_status(data_model_id, self.db)
        if current_status != "DISABLED":
            data_model_table.set_data_model_status("DISABLED", data_model_id, self.db)
        else:
            raise UserInputError("The data model was already disabled")


class EnableDataset(UseCase):
    def __init__(self, db: SQLiteDB) -> None:
        self.db = db

    def execute(self, dataset_code, data_model_code, data_model_version) -> None:
        datasets_table = DatasetsTable()
        data_model_table = DataModelTable()
        data_model_id = data_model_table.get_data_model_id(
            data_model_code, data_model_version, self.db
        )
        dataset_id = datasets_table.get_dataset_id(dataset_code, data_model_id, self.db)
        current_status = datasets_table.get_dataset_status(dataset_id, self.db)
        if current_status != "ENABLED":
            datasets_table.set_dataset_status("ENABLED", dataset_id, self.db)
        else:
            raise UserInputError("The dataset was already enabled")


class DisableDataset(UseCase):
    def __init__(self, db: SQLiteDB) -> None:
        self.db = db

    def execute(self, dataset_code, data_model_code, data_model_version) -> None:
        datasets_table = DatasetsTable()
        data_model_table = DataModelTable()
        data_model_id = data_model_table.get_data_model_id(
            data_model_code, data_model_version, self.db
        )
        dataset_id = datasets_table.get_dataset_id(dataset_code, data_model_id, self.db)
        current_status = datasets_table.get_dataset_status(dataset_id, self.db)
        if current_status != "DISABLED":
            datasets_table.set_dataset_status("DISABLED", dataset_id, self.db)
        else:
            raise UserInputError("The dataset was already disabled")


class TagDataModel(UseCase):
    def __init__(self, db: SQLiteDB) -> None:
        self.db = db

    def execute(self, code, version, tag) -> None:
        data_model_table = DataModelTable()
        data_model_id = data_model_table.get_data_model_id(code, version, self.db)
        properties = Properties(
            data_model_table.get_data_model_properties(data_model_id, self.db)
        )
        properties.add_tag(tag)
        data_model_table.set_data_model_properties(
            properties.properties, data_model_id, self.db
        )


class UntagDataModel(UseCase):
    def __init__(self, db: SQLiteDB) -> None:
        self.db = db

    def execute(self, code, version, tag) -> None:
        data_model_table = DataModelTable()
        data_model_id = data_model_table.get_data_model_id(code, version, self.db)
        properties = Properties(
            data_model_table.get_data_model_properties(data_model_id, self.db)
        )
        properties.remove_tag(tag)
        data_model_table.set_data_model_properties(
            properties.properties, data_model_id, self.db
        )


class AddPropertyToDataModel(UseCase):
    def __init__(self, db: SQLiteDB) -> None:
        self.db = db

    def execute(self, code, version, key, value, force) -> None:
        data_model_table = DataModelTable()
        data_model_id = data_model_table.get_data_model_id(code, version, self.db)
        properties = Properties(
            data_model_table.get_data_model_properties(data_model_id, self.db)
        )
        properties.add_property(key, value, force)
        data_model_table.set_data_model_properties(
            properties.properties, data_model_id, self.db
        )


class RemovePropertyFromDataModel(UseCase):
    def __init__(self, db: SQLiteDB) -> None:
        self.db = db

    def execute(self, code, version, key, value) -> None:
        data_model_table = DataModelTable()
        data_model_id = data_model_table.get_data_model_id(code, version, self.db)
        properties = Properties(
            data_model_table.get_data_model_properties(data_model_id, self.db)
        )
        properties.remove_property(key, value)
        data_model_table.set_data_model_properties(
            properties.properties, data_model_id, self.db
        )


class TagDataset(UseCase):
    def __init__(self, db: SQLiteDB) -> None:
        self.db = db

    def execute(self, dataset_code, data_model_code, data_model_version, tag) -> None:
        datasets_table = DatasetsTable()
        data_model_table = DataModelTable()
        data_model_id = data_model_table.get_data_model_id(
            data_model_code, data_model_version, self.db
        )
        dataset_id = datasets_table.get_dataset_id(dataset_code, data_model_id, self.db)
        properties = Properties(
            datasets_table.get_dataset_properties(data_model_id, self.db)
        )
        properties.add_tag(tag)
        datasets_table.set_dataset_properties(
            properties.properties, dataset_id, self.db
        )


class UntagDataset(UseCase):
    def __init__(self, db: SQLiteDB) -> None:
        self.db = db

    def execute(self, dataset, data_model_code, version, tag) -> None:
        datasets_table = DatasetsTable()
        data_model_table = DataModelTable()
        data_model_id = data_model_table.get_data_model_id(
            data_model_code, version, self.db
        )
        dataset_id = datasets_table.get_dataset_id(dataset, data_model_id, self.db)
        properties = Properties(
            datasets_table.get_dataset_properties(data_model_id, self.db)
        )
        properties.remove_tag(tag)
        datasets_table.set_dataset_properties(
            properties.properties, dataset_id, self.db
        )


class AddPropertyToDataset(UseCase):
    def __init__(self, db: SQLiteDB) -> None:
        self.db = db

    def execute(self, dataset, data_model_code, version, key, value, force) -> None:
        datasets_table = DatasetsTable()
        data_model_table = DataModelTable()
        data_model_id = data_model_table.get_data_model_id(
            data_model_code, version, self.db
        )
        dataset_id = datasets_table.get_dataset_id(dataset, data_model_id, self.db)
        properties = Properties(
            datasets_table.get_dataset_properties(data_model_id, self.db)
        )
        properties.add_property(key, value, force)
        datasets_table.set_dataset_properties(
            properties.properties, dataset_id, self.db
        )


class RemovePropertyFromDataset(UseCase):
    def __init__(self, db: SQLiteDB) -> None:
        self.db = db

    def execute(self, dataset, data_model_code, version, key, value) -> None:
        datasets_table = DatasetsTable()
        data_model_table = DataModelTable()
        data_model_id = data_model_table.get_data_model_id(
            data_model_code, version, self.db
        )
        dataset_id = datasets_table.get_dataset_id(dataset, data_model_id, self.db)
        properties = Properties(
            datasets_table.get_dataset_properties(data_model_id, self.db)
        )
        properties.remove_property(key, value)
        datasets_table.set_dataset_properties(
            properties.properties, dataset_id, self.db
        )


class ListDataModels(UseCase):
    def __init__(self, db: SQLiteDB) -> None:
        self.db = db

    def execute(self) -> None:
        data_model_table = DataModelTable()
        data_model_row_columns = ["data_model_id", "code", "version", "label", "status"]
        data_model_rows = data_model_table.get_data_models(
            db=self.db, columns=data_model_row_columns
        )
        dataset_count_by_data_model_id = {
            data_model_id: dataset_count
            for data_model_id, dataset_count in data_model_table.get_dataset_count_by_data_model_id(
                self.db
            )
        }
        data_models_info = [
            list(row) + [dataset_count_by_data_model_id.get(row[0], 0)]
            for row in data_model_rows
        ]

        if not data_models_info:
            print("There are no data models.")
            return

        df = pd.DataFrame(data_models_info, columns=data_model_row_columns + ["count"])
        print(df)


class ListDatasets(UseCase):
    def __init__(self, sqlite_db: SQLiteDB, monetdb: MonetDB) -> None:
        self.sqlite_db = sqlite_db
        self.monetdb = monetdb

    def execute(self) -> None:
        dataset_table = DatasetsTable()
        dataset_row_columns = ["dataset_id", "data_model_id", "code", "label", "status"]
        dataset_rows = dataset_table.get_datasets(
            self.sqlite_db, columns=dataset_row_columns
        )
        datasets_info = []

        for row in dataset_rows:
                datasets_info.append(list(row))

        if not datasets_info:
            print("There are no datasets.")
            return

        df = pd.DataFrame(datasets_info, columns=dataset_row_columns)
        print(df)


class Cleanup(UseCase):
    def __init__(self, sqlite_db: SQLiteDB, monetdb: MonetDB) -> None:
        self.sqlite_db = sqlite_db
        self.monetdb = monetdb

    def execute(self) -> None:
        data_model_table = DataModelTable()
        data_model_rows = data_model_table.get_data_models(
            self.sqlite_db, columns=["code", "version"]
        )

        for code, version in data_model_rows:
            DeleteDataModel(sqlite_db=self.sqlite_db, monetdb=self.monetdb).execute(
                code=code, version=version, force=True
            )


def get_data_model_fullname(code, version):
    return f"{code}:{version}"



from abc import ABC, abstractmethod
import json
from enum import Enum
from typing import List

import sqlalchemy as sql
from sqlalchemy import MetaData

from mipdb.data_frame import DATASET_COLUMN_NAME
from mipdb.databases import credentials_from_config
from mipdb.dataelements import CommonDataElement
from mipdb.exceptions import UserInputError
from mipdb.schema import Schema

RECORDS_PER_COPY = 100000


class User(Enum):
    credentials = credentials_from_config()
    executor = (
        credentials["MONETDB_LOCAL_USERNAME"]
        if credentials["MONETDB_LOCAL_USERNAME"]
        else "executor"
    )
    admin = (
        credentials["MONETDB_ADMIN_USERNAME"]
        if credentials["MONETDB_ADMIN_USERNAME"]
        else "admin"
    )
    guest = (
        credentials["MONETDB_PUBLIC_USERNAME"]
        if credentials["MONETDB_PUBLIC_USERNAME"]
        else "guest"
    )


class SQLTYPES:
    INTEGER = sql.Integer
    STRING = sql.String(255)
    FLOAT = sql.Float


STR2SQLTYPE = {"int": SQLTYPES.INTEGER, "text": SQLTYPES.STRING, "real": SQLTYPES.FLOAT}
TEMPORARY_TABLE_NAME = "temp"


class Table(ABC):
    _table: sql.Table

    @abstractmethod
    def __init__(self, schema):
        pass

    @property
    def table(self):
        return self._table

    def create(self, db):
        db.create_table(self._table)
        db.grant_select_access_rights(self._table, User.executor.value)

    def exists(self, db):
        return db.table_exists(self._table)

    def insert_values(self, values, db):
        db.insert_values_to_table(self._table, values)

    def delete(self, db):
        db.delete_table_values(self._table)

    def get_row_count(self, db):
        return db.get_row_count(self.table.fullname)

    def get_column_distinct(self, column, db):
        return db.get_column_distinct(column, self.table.fullname)

    def drop(self, db):
        db.drop_table(self._table)


metadata = MetaData()


class PrimaryDataTable(Table):
    def __init__(self):
        self._table = None

    def set_table(self, table):
        self._table = table

    @classmethod
    def from_cdes(
        cls, schema: Schema, cdes: List[CommonDataElement]
    ) -> "PrimaryDataTable":
        columns = [
            sql.Column(
                cde.code, STR2SQLTYPE[json.loads(cde.metadata)["sql_type"]], quote=True
            )
            for cde in cdes
        ]
        columns.insert(
            0,
            sql.Column(
                "row_id",
                SQLTYPES.INTEGER,
                primary_key=True,
                quote=True,
            ),
        )
        table = sql.Table(
            "primary_data",
            schema.schema,
            *columns,
        )
        new_table = cls()
        new_table.set_table(table)
        return new_table

    def get_data_count_by_dataset(self, data_model_fullname, db):
        return db.get_data_count_by_dataset(data_model_fullname)

    @classmethod
    def from_db(cls, schema: Schema, db) -> "PrimaryDataTable":
        table = sql.Table(
            "primary_data", schema.schema, autoload_with=db.get_executor()
        )
        new_table = cls()
        table.columns = [
            sql.Column(column.name, quote=True) for column in list(table.columns)
        ]
        new_table.set_table(table)
        return new_table

    def remove_dataset(self, dataset_name, schema_full_name, db):
        delete = sql.text(
            f'DELETE FROM "{schema_full_name}".primary_data '
            "WHERE dataset = :dataset_name "
        )
        db.execute(delete, dataset_name=dataset_name)


class TemporaryTable(Table):
    def __init__(self, dataframe_sql_type_per_column, db):
        columns = [
            sql.Column(name, STR2SQLTYPE[sql_type], quote=True)
            for name, sql_type in dataframe_sql_type_per_column.items()
        ]

        self._metadata = MetaData()

        self._table = sql.Table(
            TEMPORARY_TABLE_NAME,
            self._metadata,
            *columns,
            prefixes=["TEMPORARY"],
        )

        self._table.create(bind=db.get_executor())

    def create(self, db):
            db.execute(f'DROP TABLE IF EXISTS "{self.table.name}"')
            self.table.create(bind=db.get_executor())

    def validate_csv(self, csv_path, cdes_with_min_max, cdes_with_enumerations, db):
        validated_datasets = []
        offset = 2

        while True:
            self.load_csv(
                csv_path=csv_path, offset=offset, records=RECORDS_PER_COPY, db=db
            )
            offset += RECORDS_PER_COPY

            table_count = self.get_row_count(db=db)
            if not table_count:
                break

            validated_datasets = set(validated_datasets) | set(
                self.get_column_distinct(DATASET_COLUMN_NAME, db)
            )
            self._validate_enumerations_restriction(cdes_with_enumerations, db)
            self._validate_min_max_restriction(cdes_with_min_max, db)
            self.delete(db)

            # If the temp contains fewer rows than RECORDS_PER_COPY
            # that means we have read all the records in the csv and we need to stop the iteration.
            if table_count < RECORDS_PER_COPY:
                break

        return validated_datasets

    def _validate_min_max_restriction(self, cdes_with_min_max, db):
        for cde, min_max in cdes_with_min_max.items():
            min_value, max_value = min_max
            cde_invalid_values = db.execute(
                f"SELECT \"{cde}\" FROM {self.table.fullname} WHERE \"{cde}\" NOT BETWEEN '{min_value}' AND '{max_value}' "
            ).fetchone()
            if cde_invalid_values:
                raise Exception(
                    f"In the column: '{cde}' the following values are invalid: '{cde_invalid_values}'"
                )

    def load_csv(
        self,
        csv_path,
        db,
        records=None,
        offset=2,
    ):
        self._validate_csv_contains_eof(csv_path=csv_path)
        db.copy_csv_in_table(
            file_location=csv_path,
            records=records,
            offset=offset,
            table_name=self.table.name,
        )

    def _validate_csv_contains_eof(self, csv_path):
        with open(csv_path, "rb") as f:
            last_line = f.readlines()[-1]
        if not last_line.endswith(b"\n"):
            raise UserInputError(
                f"CSV:'{csv_path}' does not end with a valid EOF delimiter."
            )

    def _validate_enumerations_restriction(self, cdes_with_enumerations, db):
        for cde, enumerations in cdes_with_enumerations.items():
            cde_invalid_values = db.execute(
                f'SELECT "{cde}" from {self.table.fullname} where "{cde}" not in ({str(enumerations)[1:-1]})'
            ).fetchone()
            if cde_invalid_values:
                raise Exception(
                    f"In the column: '{cde}' the following values are invalid: '{cde_invalid_values}'"
                )

    def set_table(self, table):
        self._table = table



from mipdb.commands import disable_dataset
from mipdb.commands import disable_data_model
from mipdb.commands import enable_dataset
from mipdb.commands import enable_data_model
from mipdb.commands import init
from mipdb.commands import add_data_model
from mipdb.commands import delete_data_model
from mipdb.commands import add_dataset
from mipdb.commands import list_data_models
from mipdb.commands import list_datasets
from mipdb.commands import load_folder
from mipdb.commands import validate_dataset
from mipdb.commands import delete_dataset
from mipdb.commands import tag_dataset
from mipdb.commands import tag_data_model

__all__ = [
    "init",
    "load_folder",
    "add_data_model",
    "delete_data_model",
    "add_dataset",
    "validate_dataset",
    "delete_dataset",
    "enable_data_model",
    "disable_data_model",
    "enable_dataset",
    "disable_dataset",
    "tag_data_model",
    "tag_dataset",
    "list_data_models",
    "list_datasets",
]



from abc import ABC, abstractmethod
import json

import pandas as pd

from mipdb.exceptions import FileContentError

PANDAS_DATAFRAME_CHUNK_SIZE = 500


class Reader(ABC):
    @abstractmethod
    def read(self):
        pass


class JsonFileReader(Reader):
    def __init__(self, file) -> None:
        self.file = file

    def read(self):
        with open(self.file, "r") as stream:
            try:
                return json.load(stream)
            except json.JSONDecodeError as exc:
                orig_msg = exc.args[0]
                raise FileContentError(f"Unable to decode json file. {orig_msg}")


class CSVDataFrameReader:
    def __init__(self, file, dataframe_chunk_size=None) -> None:
        self.file = file
        self.dataframe_chunk_size = dataframe_chunk_size
        if not self.dataframe_chunk_size:
            self.dataframe_chunk_size = PANDAS_DATAFRAME_CHUNK_SIZE

    def get_reader(self):
        return pd.read_csv(
            self.file,
            dtype=object,
            chunksize=self.dataframe_chunk_size,
        )
